# -*- coding: utf-8 -*-
from phylter.backends.objects import ObjectsBackend

backends = None

if backends is None:
	backends = []

	try:
		from phylter.backends.django_backend import DjangoBackend
		backends.append(DjangoBackend)
	except ImportError:
		pass

	backends.append(ObjectsBackend)


def get_backend(o):
	for b in backends:
		if b.supports(o):
			return b()




# -*- coding: utf-8 -*-

class Condition(object):

	def __init__(self, left, right):
		self.left = left
		self.right = right

	def __eq__(self, other):
		return isinstance(other, self.__class__) and other.left == self.left and other.right == self.right

	def __repr__(self):
		return self.__str__()

class EqualsCondition(Condition):

	def __str__(self):  # pragma: nocover
		return "%s == %s" % (self.left, self.right)


class GreaterThanCondition(Condition):

	def __str__(self):  # pragma: nocover
		return "%s > %s" % (self.left, self.right)


class GreaterThanOrEqualCondition(Condition):

	def __str__(self):  # pragma: nocover
		return "%s >= %s" % (self.left, self.right)


class LessThanCondition(Condition):

	def __str__(self):  # pragma: nocover
		return "%s < %s" % (self.left, self.right)


class LessThanOrEqualCondition(Condition):

	def __str__(self):  # pragma: nocover
		return "%s <= %s" % (self.left, self.right)


class Operator(object):  # pragma: nocover
	def __init__(self, left, right):
		self.left = left
		self.right = right

	def __eq__(self, other):
		return isinstance(other, self.__class__) and other.left == self.left and other.right == self.right


class AndOperator(Operator):

	def __str__(self):  # pragma: nocover
		return "(%s) AND (%s)" % (self.left, self.right)


class OrOperator(Operator):

	def __str__(self):  # pragma: nocover
		return "(%s) OR (%s)" % (self.left, self.right)


class ConditionGroup(object):

	def __init__(self, item):
		self.item = item

	def __str__(self):
		return "(%s)" % self.item

	def __repr__(self):
		return self.__str__()

	def __eq__(self, other):
		return isinstance(other, ConditionGroup) and self.item == other.item


# -*- coding: utf-8 -*-
from django.db.models import Q
from django.db.models.manager import Manager
from django.db.models.query import QuerySet

from phylter.backends.base import Backend
from phylter.conditions import Condition, EqualsCondition, GreaterThanCondition, GreaterThanOrEqualCondition, \
	LessThanCondition, LessThanOrEqualCondition, AndOperator, OrOperator, ConditionGroup, Operator


class DjangoBackend(Backend):

	@staticmethod
	def supports(o):
		return isinstance(o, QuerySet) or isinstance(o, Manager)

	def apply(self, query, iterable):
		django_query = iterable if isinstance(iterable, QuerySet) else iterable.all()

		for o in query.query:
			django_query = self.apply_django_filter(django_query, o)

		return django_query

	def apply_django_filter(self, django_query, obj):
		return django_query.filter(self.to_q(obj))

	def to_q(self, obj):
		if isinstance(obj, Condition):
			suffix = {
				EqualsCondition: "",
				GreaterThanCondition: "__gt",
				GreaterThanOrEqualCondition: "__gte",
				LessThanCondition: "__lt",
				LessThanOrEqualCondition: "__lte",
			}[obj.__class__]

			f = "%s%s" % (obj.left, suffix)
			return Q(**{f: self.get_compatible_value(obj.right)})

		if isinstance(obj, ConditionGroup):
			return Q(self.to_q(obj.item))

		if isinstance(obj, Operator):
			if isinstance(obj, AndOperator):
				return Q(self.to_q(obj.left), self.to_q(obj.right))

			if isinstance(obj, OrOperator):
				return Q(self.to_q(obj.left)) | Q(self.to_q(obj.right))

		raise Exception("Unexpected item found in query: %s" % obj)



# -*- coding: utf-8 -*-
import re
import sys

number_re = re.compile("^-?\d+(\.\d+)?$")

def digit_or_float(s):
	return s.isdigit() or number_re.match(s) is not None

if sys.version_info[0] == 2:
	str_types = (str, unicode, bytes)
else:
	str_types = (str, bytes)


class Backend(object):

	@staticmethod
	def supports(o):  # pragma: nocover
		raise NotImplementedError

	def apply(self, query, iterable):  # pragma: nocover
		raise NotImplementedError

	def get_compatible_value(self, value, field_type=None):
		if value is None:
			return None

		value_type = type(value)

		if field_type not in str_types and field_type and value_type == field_type:
			return value

		if field_type in str_types or (field_type is None and value_type in str_types):
			s = value if isinstance(value, str_types) else str(value)

			if s[0] in ("'", '"') and s[0] == s[-1]:
				# quoted string
				return s[1:-1]

			return value

		if field_type in (int, float) or (field_type is None and value_type in (int, float)):
			if value_type in (int, float):
				return value

			if value_type in str_types and digit_or_float(value):
				return float(value)

			return value

		return value


# -*- coding: utf-8 -*-
import pyparsing

from phylter.conditions import EqualsCondition, GreaterThanCondition, LessThanCondition, GreaterThanOrEqualCondition, \
	LessThanOrEqualCondition, AndOperator, OrOperator, Condition, Operator, ConditionGroup
from phylter.query import Query

operator_signs = ('==', '!=', '>', '<', '>=', '<=')

identifier = pyparsing.Word(pyparsing.alphanums)
operator = pyparsing.oneOf(operator_signs)
value = pyparsing.quotedString | pyparsing.Word(pyparsing.alphanums)

# covers foo == bar and (foo == bar)
condition = identifier + operator + value | \
			"(" + identifier + operator + value + ")"

and_or = pyparsing.oneOf(['and', 'or'], caseless=True)
andor_field_op_value = and_or + condition

grouped_andor_field_op_value = and_or + "(" + condition + pyparsing.Optional(pyparsing.OneOrMore(andor_field_op_value)) + ")"

pattern = condition + \
		  pyparsing.Optional(pyparsing.OneOrMore(andor_field_op_value | grouped_andor_field_op_value))


class ConsumableIter(object):
	def __init__(self, iterable):
		self.iterable = iterable or []
		self.length = len(self.iterable)
		self.pos = 0

	@property
	def remaining(self):
		return self.length - self.pos

	@property
	def has_more(self):
		return self.remaining > 0

	@property
	def current(self):
		if self.pos > self.length-1:
			return None

		return self.iterable[self.pos]

	@property
	def next(self):
		if self.pos >= self.length-1:
			return None
		return self.iterable[self.pos+1]

	def consume(self, length=1):
		if length is None or length < 0 or length > self.length:
			raise ValueError("'length' argument must be 0 <= length")

		if length > self.remaining:
			raise ValueError("Cannot consume more than %s remaining elements" % self.remaining)

		if length == 0:
			return None

		if length == 1:
			element = self.iterable[self.pos]
			self.pos += length
			return element

		elements = self.iterable[self.pos:self.pos+length]
		self.pos += length
		return elements

	def __len__(self):
		return self.length

	def __contains__(self, item):
		return item in self.iterable[self.pos:]

	def __getitem__(self, item):
		return ConsumableIter(self.iterable.__getitem__(item))

	def __iter__(self):
		return self.iterable.__iter__()

	def __eq__(self, other):
		return isinstance(other, ConsumableIter) and self.iterable[self.pos:] == other.iterable[other.pos:] or \
				isinstance(other, list) and self.iterable[self.pos:] == other

class Parser(object):

	def __init__(self, *args, **kwargs):
		pass

	def parse(self, query):
		chunks = ConsumableIter(pattern.parseString(query, parseAll=True))
		return self.build_query(chunks)

	def find_group_end(self, consumable):
		openings = 1

		for i, item in enumerate(consumable):
			if item == ')':
				if openings <= 1:
					return i
				else:
					i -= 1
			elif item == '(':
				i += 1

		raise Exception("Unbalanced parenthesis")

	def build_query(self, consumable):
		l = []

		# parse groups first
		while consumable.has_more:
			if consumable.current == '(':
				consumable.consume()  # consume (
				end = self.find_group_end(consumable[consumable.pos:])
				sub = consumable[consumable.pos:consumable.pos+end] # extract group content
				consumable.consume(end+1) # consume the whole group
				sub_query = self.build_query(sub) # parse the group
				if len(sub_query.query) != 1:
					raise Exception()
				l.append(ConditionGroup(sub_query.query.iterable[0]))
			else:
				l.append(consumable.consume())

		consumable = ConsumableIter(l)
		l = []

		# replace condition tuples with *Condition instances
		while consumable.has_more:
			if consumable.current in ('and', 'or') or isinstance(consumable.current, ConditionGroup):
				l.append(consumable.consume())
			else:
				if consumable.next in operator_signs:
					left, operator, right = tuple(consumable.consume(3))
					condition = self._get_condition_class(operator)(left, right)
					l.append(condition)
				else:
					raise Exception("Unexpected tokens found: %s" % consumable.iterable[consumable.pos:])

		assert all((isinstance(x, (Condition, ConditionGroup)) or x in ('and', 'or') for x in l))

		l = ConsumableIter(l)
		# replace all operator with *Operator instances, from the lowest to the highest binding
		for op in ('or', 'and'):
			op_clazz = self._get_operator_class(op)
			l2 = []
			last = None

			while l.has_more:
				current = l.consume()

				if isinstance(current, Condition) or isinstance(current, Operator):
					last = current
				else:
					if current == op:
						last = op_clazz(last, l.consume())
					else:
						if last:
							l2.append(last)
							last = None
						l2.append(current)
			if last:
				l2.append(last)

			l = ConsumableIter(l2)

		return Query(l)

	def _get_condition_class(self, operator):
		d = {
			'==': EqualsCondition,
			'>': GreaterThanCondition,
			'<': LessThanCondition,
			'>=': GreaterThanOrEqualCondition,
			'<=': LessThanOrEqualCondition
		}
		if operator not in d:
			raise Exception("Unknown operator '%s'" % operator)
		return d[operator]

	def _get_operator_class(self, operator):
		d = {
			'and': AndOperator,
			'or': OrOperator
		}
		if operator.lower() not in d:
			raise Exception("Unknown operator '%s'" % operator)
		return d[operator]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

VERSION='0.2.3'

setup(
	name='phylter',
	version=VERSION,
	description='Library for a filter DSL in python',
	long_description='See `README.md <https://code.not-your-server.de/phylter.git>`_',
	author='Johann Schmitz',
	author_email='johann@j-schmitz.net',
	url='https://ercpe.de/projects/phylter',
	download_url='https://code.not-your-server.de/phylter.git/tags/%s.tar.gz' % VERSION,
	packages=find_packages('src'),
	package_dir={'': 'src'},
	include_package_data=True,
	zip_safe=False,
	license='GPL-3',
	install_requires=[
		'pyparsing',
	]
)



# -*- coding: utf-8 -*-
from phylter.backends import get_backend


class Query(object):

	def __init__(self, q):
		self.query = q

	def __str__(self):  # pragma: nocover
		return ' '.join([str(x) for x in self.query])

	def __repr__(self):  # pragma: nocover
		return "%s(%s)" % (self.__class__.__name__, self.__str__())

	def apply(self, iterable):
		backend = get_backend(iterable)

		return backend.apply(self, iterable)



# -*- coding: utf-8 -*-
from phylter.backends.base import Backend
from phylter.conditions import Condition, OrOperator, AndOperator, EqualsCondition, \
	GreaterThanOrEqualCondition, LessThanCondition, LessThanOrEqualCondition, GreaterThanCondition, ConditionGroup, \
	Operator


class ObjectsBackend(Backend):

	@staticmethod
	def supports(o):
		return True

	def apply(self, query, iterable):
		for item in iterable:
			if self.matches(query, item):
				yield item

	def matches(self, query, item):
		return all((self.eval_op(x, item) for x in query.query))

	def eval_op(self, op, item):
		if isinstance(op, Condition):
			left_value = getattr(item, op.left)
			right_value = self.get_compatible_value(op.right, type(left_value))

			return {
				EqualsCondition: lambda a, b: a == b,
				GreaterThanCondition: lambda a, b: a > b,
				GreaterThanOrEqualCondition: lambda a, b: a >= b,
				LessThanCondition: lambda a, b: a < b,
				LessThanOrEqualCondition: lambda a, b: a <= b,
			}[op.__class__](left_value, right_value)

		if isinstance(op, ConditionGroup):
			return self.eval_op(op.item, item)

		if isinstance(op, Operator):
			if isinstance(op, AndOperator):
				return self.eval_op(op.left, item) and self.eval_op(op.right, item)

			if isinstance(op, OrOperator):
				return self.eval_op(op.left, item) or self.eval_op(op.right, item)

		raise Exception("Unexpected item found in query: %s" % op)
